(function()
{
    var host;
    var HostId = 'com.smarttech.mathediting';
    var hostTopLeftOffset = 22;
    var defaultAnnWidth = 300;
    var annWidth = defaultAnnWidth;
    var defaultHostWidth = 343;
    var hostWidth = defaultHostWidth;
    var defaultAnnHeight = 380;
    var annHeight = defaultAnnHeight;
    var defaultHostHeight = 401;
    var hostHeight = defaultHostHeight;
    var pageWidth = 800;
    var pageHeight = 898;
    var deltaWidth = 120;
    var deltaHeight = 50;
    var extentionThreshold = 100;
    var visiblePage = {
        x: 0,
        y: 0,
        width: pageWidth,
        height: pageHeight
    };
    var omittedInks = [];
    var isEditing = false;
    var clock = new Date(2000, 0, 1);
    var inkFinishedTime = clock.getTime();
    var isInkFinished = true;
    var inkRecognitionPause = 500;
    var heartPulse = 500;
    var inkRecognizedBuffer = 30;
    var accepted = false;
    var imageRecognized;
    var mathObj = {mathML: undefined, inkObjs: undefined};
    var equationId;
    var isImageMatched = false;
    var isInserting = false;
    var mathInksObjs = [];
    var inkTypes = ['group', 'ink', 'calligraphicInk', 'creativeInk', 'highlighter', 'straightLine'];
    var dragging = false;
    var updateInks = false;
    var reBindToHost = false;
    var loadedMathInksObjs = [];
    var widgetSizeKey = 'MIE_WidgetSize';
    var DefaultWidgetSize = {'width': defaultAnnWidth, 'height': defaultAnnHeight};
    var inkIdsKey = 'MIE_InkIds';
    var equationIdKey = 'MIE_EquationId';
    var isEditingKey = 'MIE_IsEditing';
    var stayAlive = true;
    var loadingMathInk = false;

    function updateWidget()
    {
        host = NB.getHostObject();
        annWidth = host.width + deltaWidth;
        hostWidth = parseInt($('.editor').css('width')) + deltaWidth;
        hostHeight = parseInt($('.editor').css('height'));
        var rightEdge = visiblePage.x + visiblePage.width;
        if (host.x + hostTopLeftOffset + hostWidth > rightEdge)
        {
            hostWidth = rightEdge - host.x - hostTopLeftOffset;
            annWidth = hostWidth + 2 * hostTopLeftOffset;
        }
       NB.undomanager.setAutoRecord(false);
        $('.editor').css('width', hostWidth);
        host.width = annWidth;
       NB.undomanager.setAutoRecord(true);

       updatePersistWidgetSizeData(hostWidth, hostHeight);
    }

    function updateWidgetVertically()
    {
        host = NB.getHostObject();
        annHeight = host.height + deltaHeight;
        hostWidth = parseInt($('.editor').css('width'));
        hostHeight = parseInt($('.editor').css('height')) + deltaHeight;

        NB.undomanager.setAutoRecord(false);
        $('.editor').css('height', hostHeight);
        host.height = annHeight;
        NB.undomanager.setAutoRecord(true);
        updatePersistWidgetSizeData(hostWidth, hostHeight);
    }

    function updateVisiblePage()
    {
        var area = NB.page.getVisibleArea();
        if (area.x < 0)
        {
            visiblePage.x = 0;
            visiblePage.width = pageWidth;
        }
        else
        {
            visiblePage.x = area.x;
            visiblePage.width = area.width;
        }

        visiblePage.y = area.y;
        visiblePage.height = area.height;
    }

    function resetWidget(keepInks)
    {
        NB.undomanager.setAutoRecord(false);
        if (!isDefined(keepInks))
        {
            var inkObjs = getInkObjects();
            if (isDefined(inkObjs))
            {
                for (var j = 0; j < inkObjs.length; j++)
                {
                    inkObjs[j].deleteObject();
                    inkObjs[j].deleteHistory();
                }
            }
         }
         NB.batch(function()
         {
            stayAlive = false;
         // DO NOT CHANGE the order of following statements !
            host.x = -pageWidth;
            host.y = 0;
            annWidth = defaultAnnWidth;
            host.width = defaultHostWidth;
            hostWidth = defaultHostWidth;
            $('.editor').css('width', defaultAnnWidth);
            $('.editor').css('height', defaultAnnHeight);
            $('.equationoutput .equation').css('background-image', 'none');
            isEditing = false;
         });

        updatePersistWidgetSizeData(defaultAnnWidth, defaultAnnHeight);

        imageRecognized = null;
        mathObj.mathML = null;
        mathObj.inkObjs = null;
        imageRecognized = undefined;
        mathObj.mathML = undefined;
        mathObj.inkObjs = undefined;
        equationId = undefined;
        isImageMatched = false;
        isInserting = false;
        $('input#acceptId').css('opacity', 1.0);
        $('#waitspinner').css('visibility', 'hidden');
        setTextRecognition(false);
        updateInks = true;
        reBindToHost = false;
        for (var i = 0; i < mathInksObjs.length; i++)
        {
            mathInksObjs[i].deleteObject();
            mathInksObjs[i].deleteHistory();
        }
        for (i = 0; i < loadedMathInksObjs.length; i++)
        {
            loadedMathInksObjs[i].deleteObject();
            loadedMathInksObjs[i].deleteHistory();
        }
        NB.undomanager.setAutoRecord(true);
        mathInksObjs = [];
        loadedMathInksObjs = [];
        updatePersistEditModeData();
        // remove all history of the widget - this will remove undo history of the widget and leave user with recognized equation history
        NB.undomanager.removeHistory(host.guid);
    }

    function preShowWidget()
    {
        stayAlive = true;
        findOmittedInks();
        setTextRecognition(true);
    }

    function hideWidget()
    {
        NB.undomanager.setAutoRecord(false);
        NB.batch(function()
        {
            host.x = -pageWidth;
            host.y = 0;
        });
        NB.undomanager.setAutoRecord(true);
    }

    function findOmittedInks()
    {
        omittedInks = [];
        var objs = NB.page.getObjectsAsArray();
        omittedInks = objs.filter(function(elm) {
            if (isInk(elm))
            {
                return true;
            }
        });
        //console.log('Number of omitted: ' +  JSON.stringify(omittedInks.length));
    }

    function getInkObjects()
    {
        var inkObjs = [];
        if (updateInks)
        {
            var objs = NB.page.getObjectsAsArray();
            inkObjs = objs.filter(function(elm) {
                if (isInk(elm))
                {
                    if (isCenterOnTextArea(elm))
                    {
                        return omittedInks.every(function(omittedElm) {
                            return elm.guid != omittedElm.guid;
                        });
                    }
                }
            });

            if(inkObjs.length > 0)
            {
                updateInks = false;
                reBindToHost = true;
            }
            return inkObjs;
        }
        else
        {
            return mathInksObjs;
        }
    }

    function isCenterOnTextArea(ann)
    {
        if (isDefined(ann))
        {
            if (isInsideTextArea(ann.x + ann.width / 2, ann.y + ann.height / 2, 0, 0, 0))
            {
                return true;
            }
        }
        return false;
    }

    function isOnWidget(ann)
    {
        host = NB.getHostObject();
        if (host.contains(ann.x, ann.y) && host.contains(ann.x + ann.width, ann.y + ann.height))
        {
            return true;
        }
        return false;
    }

    function isOnTextArea(ann)
    {
        return isInsideTextArea(ann.x, ann.y, ann.height, ann.width, inkRecognizedBuffer);
    }

    function isInsideTextArea(x, y, height, width, buffer)
    {
        host = NB.getHostObject();
        var textArea = {};
        textArea.left = host.x + hostTopLeftOffset - buffer;
        textArea.top = host.y + hostTopLeftOffset - buffer;
        textArea.right = textArea.left + hostWidth;
        textArea.bottom = textArea.top + hostHeight + 2 * buffer;

        if (x > textArea.left && x < textArea.right &&
            y > textArea.top && y < textArea.bottom &&
            (x + width) > textArea.left && (x + width) < textArea.right &&
            (y + height) > textArea.top && (y + height) < textArea.bottom)
        {
            return true;
        }
        return false;
    }

    function isInk(inkObj)
    {
        return inkTypes.some(function(mathType) {
            return inkObj.type == mathType;
        });
    }

    function isInkChanged(inkObjs)
    {
        if (isDefined(inkObjs) && inkObjs.length === 0 || !isDefined(mathObj.inkObjs))
        {
            return true;
        }

        if (inkObjs.length != mathObj.inkObjs.length)
        {
            return true;
        }

        var diff = inkObjs.filter(function(elm) {
            if (isInk(elm))
            {
                return mathObj.inkObjs.every(function(mathElm) {
                    return elm.guid != mathElm.guid;
                });
            }
        });

        if (diff.length === 0)
        {
            return false;
        }
        else
        {
            return true;
        }
    }

    function onAccept()
    {
        var inkObjs = getInkObjects();
        if (isDefined(inkObjs) && inkObjs.length > 0)
        {

            if (!isInkChanged(inkObjs) && isImageMatched)
            {
                insertEquation();
                console.log('Insert Equation WITHOUT Image conversion !');
                return;
            }

            NB.hwr.resetHWERSession();
            console.log('resetHWERSession() called!');
            mathObj.mathML = undefined;
            mathObj.inkObjs = undefined;
            accepted = true;
            $('input#acceptId').css('opacity', 0.2);
            $('#waitspinner').css('visibility', 'visible');
            mathInksObjs = inkObjs;
            NB.hwr.getHWEREquation(inkObjs, HWEREquationCallBack);
            NB.hwr.resetHWERSession();
        }
        else
        {
            resetWidget();
        }
    }

    function setTextRecognition(ready)
    {
        if (ready)
        {
            clock = new Date();
            inkFinishedTime = clock.getTime();
            isInkFinished = true;
        }
        else
        {
            clock = new Date(2100, 0, 1);
            inkFinishedTime = clock.getTime();
            isInkFinished = false;
        }
    }

    function HWERControlerAsync($scope)
    {
       //  clock = new Date();
       //  var now = clock.getTime();
       //  var pause = parseInt(now) - parseInt(inkFinishedTime);
       // if (!dragging && isInkFinished && (pause > inkRecognitionPause))
        if (!dragging)
        {
                var inkObjs = getInkObjects();
                if (isDefined(inkObjs) && inkObjs.length > 0)
                {
                        NB.hwr.getHWEREquation(inkObjs, HWEREquationCallBack);
                        saveInksOnWidget(inkObjs);
                }
                else
                {
                    NB.hwr.resetHWERSession();
                    mathObj.mathML = undefined;
                    mathObj.inkObjs = undefined;
                    $('.equationoutput .equation').css('background-image', 'none');
                }

                if (reBindToHost)
                {
                    unbindToHost(mathInksObjs);
                    mathInksObjs = inkObjs;
                    bindToHost(mathInksObjs);
                    reBindToHost = false;
                }
        }
        //console.log('inkingEvent finished: ' +  JSON.stringify(isInkFinished) + ' pause: ' + pause );
    }

    var HWEREquationCallBack = function(evt)
    {
        console.log('HWEREquationCallBack() called from HWERControlerAsync()!');
        if (evt === null)
        {
            console.log('NB.hwr.getHWEREquation failed!');
            return;
        }

        if (isDefined(evt.eventData.equation) && evt.eventData.equation.length > 0)
        {
            var mathMLResult = evt.eventData.equation;
            if (mathMLResult != mathObj.mathML)
            {
                var color;
                if (isDefined(mathInksObjs[0].stroke))
                {
                    color = mathInksObjs[0].stroke.color;
                }

                if (!isDefined(color))
                {
                    color = [0, 0, 0];
                }
                NB.hwr.getEquationImageAsync(mathMLResult, color);
                mathObj.mathML = mathMLResult;
                mathObj.inkObjs = mathInksObjs;
                isImageMatched = false;
                //console.log("Updated mathMLResult: " + JSON.stringify(mathMLResult));
            }
        }
    };

    function updatePreview(imagePath, previewWidth, previewHeight, imageWidth, imageHeight)
    {

        var updateHeight = previewHeight;
        var updateWidth = imageWidth / imageHeight * previewHeight;
        if (updateWidth > previewWidth)
        {
            updateHeight = imageHeight / imageWidth * previewWidth;
            updateWidth = previewWidth;
        }

        var size = updateWidth + 'px' + ' ' + updateHeight + 'px';

        if (imagePath === '')
        {
            $('.equationoutput .equation').css('background-size', size).css('background-image', 'none');
        }
        else
        {
            var image = 'url(\'smttmp^' + imagePath + '\') no-repeat';
            $('.equationoutput .equation').css('background', image).css('background-size', size);
        }
    }

    function updateImage(eventData)
    {
        host = NB.getHostObject();
        if ((host.x + host.width) >= 0)
        {
            if (isAllPropDefined(eventData))
            {
                var imagePath = eventData.imgPath;
                var width = parseInt($('.equationoutput .equation').css('width'));
                var height = parseInt($('.equationoutput .equation').css('height'));
                var imageWidth = eventData.width;
                var imageHeight = eventData.height;

                updatePreview(imagePath, width, height, imageWidth, imageHeight);

                if (eventData.mathML == mathObj.mathML)
                {
                    isImageMatched = true;
                    console.log('Image updated matches the mathML!');
                    //console.log("Matched mathMLResult: " + JSON.stringify(mathObj.mathML));
                }
                console.log('Math Image Updated !');
            }
        }
        else
        {
            $('.equationoutput .equation').css('background-image', 'none');
        }
    }

    function progress()
    {
        $('.equationoutput .equation').css('background-image', 'none');
            var running = false;
            if (!running) {
            running = true;
            var count = 0;
            var inter = null;
            var runProgressfunction = function() {
                count += 1;
                $('.equationoutput .equation').css('background-color', 'green').css('width', count + '%');
                if (count == 88) {
                    clearInterval(inter);
                    running = false;
                $('.equationoutput .equation').css('background-color', '#eef4f9');
                }
            };
            inter = setInterval(runProgressfunction, 30);
        }
    }

    function updateEquation()
    {
        try
        {
            if (isAllPropDefined(imageRecognized) && isAllPropDefined(mathObj))
            {
                accepted = false;
                var imageFile = imageRecognized.imgPath;
                host = NB.getHostObject();
                var annWidth = getMathInksWidth(mathObj.inkObjs);
                var annHeight = imageRecognized.height / imageRecognized.width * annWidth;
                if (!isEditing)
                {
                    // start record for recognized equation - on undo it will remove recognized annotation, on redo will bring it back
                    NB.undomanager.startRecord("create equation");
                    var id = NB.page.createEquationAndAddToPage(mathObj.mathML, imageFile, host.x, host.y, annWidth, annHeight);
                    equationId = id.Result;
                }
                var equation = NB.getObject(equationId);
                if (isDefined(equation))
                {
                    if (isEditing)
                    {
                        // create record of recognized annotation update
                        NB.undomanager.startRecord("update equation");
                    }
                    equation.setEquationInks(host.x, host.y, mathObj.inkObjs);
                    equation.setEquationWidgetSize(host.width, host.height);
                    console.log('host.width: %i  host.height: %i', host.width, host.height);
                    if (isEditing)
                    {
                        equation.setMathML(mathObj.mathML);
                        equation.setEquationImageFile(imageFile);
                        equation.width = annWidth;
                        equation.height = annHeight;
                    }
                    NB.undomanager.stopRecord();
                }
                else
                {
                    // looks like as unexpected case - finish undo/redo record if started
                    if (!isEditing)
                    {
                        NB.undomanager.stopRecord();
                    }
                }
            }
        }
        catch (err)
        {
            console.log('UpdateEquation failed equationId: %i  ImageWidth: %i', equationId, imageRecognized.width);
        }
    }

    function resetEquation()
    {
        var equation = NB.getObject(equationId);
        if (isDefined(equation))
        {
            NB.undomanager.setAutoRecord(false);
            var originalMathInksObjs = [];
            var inkIds = equation.loadEquationInks(host.x, host.y);
            for (var i = 0; i < inkIds.length; i++)
            {
                originalMathInksObjs.push(NB.getObject(inkIds[i]));
            }
            equation.setEquationInks(host.x, host.y, originalMathInksObjs);
            for (i = 0; i < originalMathInksObjs.length; i++)
            {
                originalMathInksObjs[i].deleteObject();
                originalMathInksObjs[i].deleteHistory();
            }
            NB.undomanager.setAutoRecord(true);
        }
    }

    function insertEquation()
    {
        var theEquationId;
        updateEquation();
        theEquationId = equationId;
        resetEquation();
        resetWidget();
    }

    function dismissEditing()
    {
        resetWidget();
    }

    function isAllPropDefined(obj)
    {
        if (!isDefined(obj))
        {
            return false;
        }

        for (var item in obj)
        {
            if (!isDefined(item))
            {
                return false;
            }
        }
        return true;
    }

    function getMathInksWidth(objs)
    {
        var minX = Number.MAX_VALUE;
        var maxX = - minX;

        if (objs.length == 1)
        {
            // objs[0].width is not correct when length is 1
            return NB.getObject(objs[0].guid).width;
        }

        for (var i in objs)
        {
            var left = objs[i].x;
            if (left < minX)
            {
                minX = left;
            }

            var right = objs[i].x + objs[i].width;

            if (right > maxX)
            {
                maxX = right;
            }
        }
        return maxX - minX;
    }

    function removeFromOmittedInk(inkGuids)
    {
        omittedInks = omittedInks.filter(function(elm) {
            return inkGuids.every(function(guid) {
                return elm.guid != guid;
            });
        });
    }

    function bindToHost(objs)
    {
        if (isDefined(objs) && objs.length > 0)
        {
            objs.forEach(function(obj) {
                {
                    obj.bindProperty(NB.HOSTGUID, 'deltaX', 'deltaX');
                    obj.bindProperty(NB.HOSTGUID, 'deltaY', 'deltaY');
                    obj.setSavableProperty(false);
                }
            });
            console.log('bindToHost - obj count: %i  ', objs.length);
        }
    }

    function unbindToHost(objs)
    {
        if (isDefined(objs) && objs.length > 0)
        {        
            objs.forEach(function(obj) {
                {
                    obj.unbindProperty('deltaX');
                    obj.unbindProperty('deltaY');
                    obj.setSavableProperty(true);
                }
            });
            console.log('unbindToHost - obj count: %i  ', objs.length);
        }
    } 

    function updateWidgetSize(widgetSize)
    {
        if(widgetSize.width !== DefaultWidgetSize.width)
        {
            $('.editor').css('width', widgetSize.width);
            hostWidth = widgetSize.width;
        }

        if(widgetSize.height !== DefaultWidgetSize.height)
        {
            $('.editor').css('height', widgetSize.height);
            hostHeight = widgetSize.height;
        }
    }
  
    function updatePersistWidgetSizeData(widgetWidth, widgetHeight)
    {
        var widgetSize;
        var localDataKeys = NB.persist.list("session");
        if(localDataKeys !== null && isDefined(localDataKeys) && localDataKeys.indexOf(widgetSizeKey) !== -1)
        {
            widgetSize = NB.persist.load(widgetSizeKey, "session");
        }

        if(isDefined(widgetSize))
        {
            widgetSize = JSON.parse(widgetSize);
            //console.log('updatePersistWidgetSizeData !');
            //console.log('editorWidth.: %i  editorHeight: %i', widgetWidth, widgetHeight);
            //console.log('widgetSize.width: %i  widgetSize.height: %i', widgetSize.width, widgetSize.height);            

            if(widgetSize.width !== widgetWidth)
            {
                widgetSize.width = widgetWidth;
                NB.persist.save(widgetSizeKey, widgetSize, "session");
            }

            if(widgetSize.height !== widgetHeight)
            {
                widgetSize.height = widgetHeight;
                NB.persist.save(widgetSizeKey, widgetSize, "session"); 
            }
        }
    }

    function updatePersistEditModeData()
    {
        NB.persist.save(isEditingKey, isEditing, "session");
        NB.persist.save(equationIdKey, equationId, "session");
    }

    function saveInksOnWidget(inkObjs)
    {
        var inkIds = [];
        inkObjs.forEach(function(obj) {
            inkIds.push(obj.guid);
        });
        NB.persist.save(inkIdsKey, inkIds, "session");
    }

    function setReadyForEditing(editingEquationId)
    {
        setTimeout(function() {
            NB.hwr.resetHWERSession();
            host = NB.getHostObject();
            equationId = editingEquationId;
            var eObj = NB.getObject(equationId);
            var size = eObj.getEquationWidgetSize();
            //console.log('size.width: %i  size.height: %i', size.width, size.height);
            hostWidth = size.width - hostTopLeftOffset * 2;
            hostHeight = size.height - hostTopLeftOffset * 2;
            NB.undomanager.setAutoRecord(false);
            $('.editor').css('width', hostWidth);
            host.width = size.width;
            $('.editor').css('height', hostHeight);
            host.height = size.height;
            NB.undomanager.setAutoRecord(true);

            updatePersistWidgetSizeData(hostWidth, hostHeight);
            var width = hostWidth*0.88;
            var height = parseInt($('.equationoutput .equation').css('height'));

            updatePreview('', width, height, eObj.width, eObj.height);

            var hostX = eObj.x;
            var hostY = eObj.y;

            if (hostX + host.width > pageWidth)
            {
                hostX = pageWidth - host.width;
            }
            if (hostX < 0)
            {
                hostX = 0;
            }
            if (hostY + host.height > pageHeight)
            {
                hostY = pageHeight - host.height;
            }
            if (hostY < 0)
            {
                hostY = 0;
            }
            preShowWidget();
            if(!loadingMathInk)
            {
                loadingMathInk = true;
                setTimeout(function() {
                    // add one undo record for widget positioning and created ink annotations
                    NB.undomanager.startRecord("update widget position");
                    host.moveToFront();
                    host.x = hostX;
                    host.y = hostY;
                    var inkIds = eObj.loadEquationInks(host.x, host.y);
                    NB.undomanager.stopRecord(true);
                    for (var i = 0; i < inkIds.length; i++)
                    {
                         loadedMathInksObjs.push(NB.getObject(inkIds[i]));
                    }
                    loadingMathInk = false;
                }, 300);
            }
            isEditing = true;
            updatePersistEditModeData();
        }, 100);
    }

    NB.ready(function()
    {
        var isNBMarkedDirty = NB.document.isDirty();
        host = NB.getHostObject();
        // no need undo/redo record of widget id change
        NB.undomanager.setAutoRecord(false);
        host.widget.id = HostId;
        NB.undomanager.setAutoRecord(true);
        var widgetSize;
        var inkIdsVal;
        var localDataKeys = NB.persist.list("session");
        if (localDataKeys === null)
        {
            NB.persist.save(widgetSizeKey, DefaultWidgetSize, "session");
        }
        else if(isDefined(localDataKeys))
        {
            if (localDataKeys.indexOf(widgetSizeKey) !== -1)
            {
                widgetSize = NB.persist.load(widgetSizeKey, "session");
                if(isDefined(widgetSize))
                {
                    updateWidgetSize(JSON.parse(widgetSize));
                }
            }

            if(localDataKeys.indexOf(inkIdsKey) !== -1)
            {
                inkIdsVal = NB.persist.load(inkIdsKey, "session");
            }

            var isInEditing;
            if(localDataKeys.indexOf(isEditingKey) !== -1)
            {
                isInEditing = NB.persist.load(isEditingKey, "session");
                if(isDefined(isInEditing))
                {
                    isEditing = JSON.parse(isInEditing);
                }
            }

            var equId;
            if(localDataKeys.indexOf(equationIdKey) !== -1)
            {
                equId = NB.persist.load(equationIdKey, "session");
                if(isDefined(equId))
                {
                    equationId = JSON.parse(equId);
                }
            }
        }
        
        NB.addObserver('preShowMathWidget', function(evt) {
            preShowWidget();
            console.log('preShowMathWidget');
        });

        NB.addObserver('postShowMathWidget', function(evt) {
            console.log('postShowMathWidget');
        });

       NB.addObserver('pageLeaveEvent', function(evt) {
           var isNBMarkedDirty = NB.document.isDirty();
           var eventData = {};
           eventData['com.smarttech.addon.alignment'] = {'isNBDirt': isNBMarkedDirty};
           NB.postEvent('prePageLeaveEvent', eventData);
           stayAlive = false;
           console.log('pageLeaveEvent:   alive - ' + JSON.stringify(stayAlive));

       });

       NB.addObserver('pageEnterEvent', function(evt) {
           stayAlive = true;
           console.log('pageEnterEvent:   alive - ' + JSON.stringify(stayAlive));
       });

        NB.addObserver('startMathEditing', function(evt) 
            {
                var editingEquationId = evt.eventData.guidMath;

                if( editingEquationId != equationId)
                {
                    resetWidget();
                    setReadyForEditing(editingEquationId);
                }
            }
        );

        NB.addObserver('EquationImageEvent', function(evt) {
            imageRecognized = evt.eventData;
            if (accepted && imageRecognized.mathML == mathObj.mathML)
            {
                insertEquation();
                console.log('Insert Equation WITH matched Image !');
            }
            else
            {
                updateImage(evt.eventData);
            }
        });

        NB.addObserver('inkingEvent', function(evt)
        {
            // console.log('inkingEvent inking: ' +  JSON.stringify(isInkFinished));
            var x = evt.eventData.x;
            var y = evt.eventData.y;
            if (isInsideTextArea(x, y, 0, 0, 0))
            {
                isInkFinished = false;
                var threshold = host.x + annWidth - extentionThreshold;
                if (x > threshold)
                {
                    updateWidget();
                }
                threshold = host.y + annHeight - extentionThreshold;
                if (y > threshold)
                {
                    updateWidgetVertically();
                }
            }

        });

        // NB.addObserver('inkingFinishedEvent', function(evt)
        // {
        //     if (host.contains(evt.eventData.x, evt.eventData.y))
        //     {
        //         clock = new Date();
        //         inkFinishedTime = clock.getTime();
        //         isInkFinished = true;
        //         //console.log('inkingEvent finished: ' +  JSON.stringify(isInkFinished));
        //     }
        // });

         NB.addObserver('dndStartEvent', function(evt)
         {
            if (isDefined(evt.eventData[host.guid]))
            {
                dragging = true;
            }

            for (var key in evt.eventData) {
                var obj = NB.getObject(key);
                if (isInk(obj) && isCenterOnTextArea(obj))
                {
                    updateInks = true;
                }
            }
         });

        NB.addObserver('dndDropEvent', function(evt) {

            if (isDefined(evt.eventData[host.guid]))
            {
                dragging = false;
                updateInks = true;
            }
            else
            {
                var droppedInks = [];
                var guidEquals = function (elm) {
                    return elm.guid == this.guid;
                };
                var guidNotEquals = function(elm) {
                    return elm.guid != this.guid;
                };
                for (var key in evt.eventData) {
                    var obj = NB.getObject(key);
                    if (isInk(obj))
                    {
                        var wasInside = false;
                        var thisObj = {guid: key};
                        if(isDefined(mathInksObjs))
                        {
                            wasInside = mathInksObjs.some(guidEquals, thisObj);
                        }

                        var wasOutside = omittedInks.some(guidEquals, thisObj);

                        if(isCenterOnTextArea(obj))
                        {
                            if(wasOutside)
                            {
                                obj.moveToFront();
                                droppedInks.push(key);
                            }
                            updateInks = true;
                        }
                        else
                        {   
                            if(wasInside)
                            {
                               updateInks = true;
                               obj.unbindProperty('deltaX');
                               obj.unbindProperty('deltaY');
                               thisObj = {guid: obj.guid};
                               mathInksObjs = mathInksObjs.filter(guidNotEquals, thisObj);
                            }
                        }
                    }
                }

                removeFromOmittedInk(droppedInks);
            }

            if (updateInks)
            {
                NB.hwr.resetHWERSession();
                mathObj.mathML = undefined;
                mathObj.inkObjs = undefined;
            }

        });

        NB.addObserver('annotationsUngroupedEvent', function(evt) {
            var obj = NB.getObject(evt.eventData.ungroupedGroup);
            if (isInk(obj) && isCenterOnTextArea(obj))
            {
//               console.log('annotationsUngroupedEvent - obj guid:', evt.eventData.ungroupedGroup);
            }

        });

        NB.addObserver('annotationAddedEvent', function(evt) {
            var obj = NB.getObject(evt.eventData.guid);
            if (isInk(obj) && isCenterOnTextArea(obj))
            {
                updateInks = true;
//                console.log('annotationAddedEvent - obj guid:', evt.eventData.guid);
            }

        });

        NB.addObserver('annotationDeletedEvent', function(evt) {
            var obj = NB.getObject(evt.eventData.guid);
            if (isInk(obj) && isCenterOnTextArea(obj))
            {
                updateInks = true;
//                console.log('annotationDeletedEvent - obj guid:', evt.eventData.guid); 
            }

        });

        NB.addObserver('annotationsErasedEvent', function(evt) {
            var guids = evt.eventData.after;

            if(!isDefined(guids))
            {
                guids = evt.eventData.before;
            }

            if(isDefined(guids))
            {
                guids.forEach(function(guid) {
                    var obj = NB.getObject(guid);
                    if (isInk(obj) && isCenterOnTextArea(obj))
                    {
                        updateInks = true;
                    }
                });
            }

        });

        // need to initialize angularjs manually so it initializes after NBSDK is ready
        angular.element(document).ready(function()
        {
            var MathEditingApp = angular.module('MathEditingApp', ['ngAnimate']);

            // defining angular controller
            MathEditingApp.controller('MathController', ['$scope', '$timeout',
                function($scope, $timeout)
                {
                    $scope.isInserting = isInserting;

                    // $scope.down = function()
                    // {
                        // $('input#acceptId').css('opacity', 0.2);
                    // };

                    // $scope.up = function()
                    // {
                        // $('input#acceptId').css('opacity', 1.0);
                    // };
                    $scope.acceptInk = function()
                    {
                        onAccept();
                    };

                    $scope.acceptRecognized = function()
                    {
                        insertEquation();
                        console.log('Equation inserted from preview!');
                    };

                    $scope.rejectInk = function()
                    {
                        dismissEditing();
                    };

                    var poll = function()
                    {
                        $timeout(function()
                        {
                            if(stayAlive)
                            {
                                HWERControlerAsync($scope);
                            }
                            poll();
                            console.log('Heartbeat:  alive - ' + JSON.stringify(stayAlive));
                        }, heartPulse);
                    };
                    poll();
                }
            ]);

            angular.bootstrap(document, ['MathEditingApp']);
        });
        var eventData = {};
        eventData['com.smarttech.addon.alignment'] = {'widgetId': host.guid, 'isNBDirt': isNBMarkedDirty};
        NB.postEvent('mathEditingReady', eventData);
        findOmittedInks();
        var inkIds = [];
        if (isDefined(inkIdsVal))
        {
            inkIds = JSON.parse(inkIdsVal);
            if (inkIds.length > 0)
            {
                removeFromOmittedInk(inkIds);
            }
        }
        updateInks = true;
        NB.hwr.resetHWERSession();
    });
})();
